/* Copyright (c) 2008 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.watto.program.android.sync.miso;

import java.io.IOException;
import java.util.Collection;


/**
 * A {@code UnicodeEscaper} that escapes some set of Java characters using
 * the URI percent encoding scheme. The set of safe characters (those which
 * remain unescaped) can be specified on construction.
 *
 * <p>For details on escaping URIs for use in web pages, see section 2.4 of
 * <a href="http://www.ietf.org/rfc/rfc3986.txt">RFC 3986</a>.
 *
 * <p>In most cases this class should not need to be used directly. If you
 * have no special requirements for escaping your URIs, you should use either
 * {@link CharEscapers#uriEscaper()} or
 * {@link CharEscapers#uriEscaper(boolean)}.
 *
 * <p>When encoding a String, the following rules apply:
 * <ul>
 * <li>The alphanumeric characters "a" through "z", "A" through "Z" and "0"
 * through "9" remain the same.
 * <li>Any additionally specified safe characters remain the same.
 * <li>If {@code plusForSpace} was specified, the space character " " is
 * converted into a plus sign "+".
 * <li>All other characters are converted into one or more bytes using UTF-8
 *     encoding and each byte is then represented by the 3-character string
 *     "%XY", where "XY" is the two-digit, uppercase, hexadecimal representation
 *     of the byte value.
 * </ul>
 *
 * <p>RFC 2396 specifies the set of unreserved characters as "-", "_", ".", "!",
 * "~", "*", "'", "(" and ")". It goes on to state:
 *
 * <p><i>Unreserved characters can be escaped without changing the semantics
 * of the URI, but this should not be done unless the URI is being used
 * in a context that does not allow the unescaped character to appear.</i>
 *
 * <p>For performance reasons the only currently supported character encoding of
 * this class is UTF-8.
 *
 * <p><b>Note</b>: This escaper produces uppercase hexidecimal sequences. From
 * <a href="http://www.ietf.org/rfc/rfc3986.txt">RFC 3986</a>:<br>
 * <i>"URI producers and normalizers should use uppercase hexadecimal digits
 * for all percent-encodings."</i>
 *
 * 
 */
public class PercentEscaper {

  /**
   * A string of safe characters that mimics the behavior of
   * {@link java.net.URLEncoder}.
   *
   */
  public static final String SAFECHARS_URLENCODER = "-_.*";

  /** The amount of padding (chars) to use when growing the escape buffer. */
  private static final int DEST_PAD = 32;

  /**
   * A string of characters that do not need to be encoded when used in URI
   * path segments, as specified in RFC 3986. Note that some of these
   * characters do need to be escaped when used in other parts of the URI.
   */
  public static final String SAFEPATHCHARS_URLENCODER = "-_.!~*'()@:$&,;=";

  /**
   * A string of characters that do not need to be encoded when used in URI
   * query strings, as specified in RFC 3986. Note that some of these
   * characters do need to be escaped when used in other parts of the URI.
   */
  public static final String SAFEQUERYSTRINGCHARS_URLENCODER = "-_.!~*'()@:$,;/?:";

  // In some uri escapers spaces are escaped to '+'
  private static final char[] URI_ESCAPED_SPACE = {'+'};

  private static final char[] UPPER_HEX_DIGITS = "0123456789ABCDEF".toCharArray();

  /**
   * If true we should convert space to the {@code +} character.
   */
  private final boolean plusForSpace;

  /**
   * An array of flags where for any {@code char c} if {@code safeOctets[c]} is
   * true then {@code c} should remain unmodified in the output. If
   * {@code c > safeOctets.length} then it should be escaped.
   */
  private final boolean[] safeOctets;

  /**
   * A thread-local destination buffer to keep us from creating new buffers.
   * The starting size is 1024 characters.  If we grow past this we don't
   * put it back in the threadlocal, we just keep going and grow as needed.
   */
  private static final ThreadLocal<char[]> DEST_TL = new ThreadLocal<char[]>() {

    @Override
    protected char[] initialValue(){
      return new char[1024];
    }
  };


  /**
   * Constructs a URI escaper with the specified safe characters and optional
   * handling of the space character.
   *
   * @param safeChars a non null string specifying additional safe characters
   *        for this escaper (the ranges 0..9, a..z and A..Z are always safe and
   *        should not be specified here)
   * @param plusForSpace true if ASCII space should be escaped to {@code +}
   *        rather than {@code %20}
   * @throws IllegalArgumentException if any of the parameters were invalid
   */
  public PercentEscaper(String safeChars, boolean plusForSpace){
    // Avoid any misunderstandings about the behavior of this escaper
    if (safeChars.matches(".*[0-9A-Za-z].*")) {
      throw new IllegalArgumentException("Alphanumeric characters are always 'safe' and should not be " + "explicitly specified");
    }
    // Avoid ambiguous parameters. Safe characters are never modified so if
    // space is a safe character then setting plusForSpace is meaningless.
    if (plusForSpace && safeChars.contains(" ")) {
      throw new IllegalArgumentException("plusForSpace cannot be specified when space is a 'safe' character");
    }
    if (safeChars.contains("%")) {
      throw new IllegalArgumentException("The '%' character cannot be specified as 'safe'");
    }
    this.plusForSpace = plusForSpace;
    this.safeOctets = createSafeOctets(safeChars);
  }


  /**
   * Returns an {@code Appendable} instance which automatically escapes all
   * text appended to it before passing the resulting text to an underlying
   * {@code Appendable}.
   *
   * <p>Unlike {@link #escape(String)} it is permitted to append arbitrarily
   * split input to this Appendable, including input that is split over a
   * surrogate pair. In this case the pending high surrogate character will not
   * be processed until the corresponding low surrogate is appended. This means
   * that a trailing high surrogate character at the end of the input cannot be
   * detected and will be silently ignored. This is unavoidable since the
   * Appendable interface has no {@code close()} method, and it is impossible to
   * determine when the last characters have been appended.
   *
   * <p>The methods of the returned object will propagate any exceptions thrown
   * by the underlying {@code Appendable}.
   *
   * <p>For well formed <a href="http://en.wikipedia.org/wiki/UTF-16">UTF-16</a>
   * the escaping behavior is identical to that of {@link #escape(String)} and
   * the following code is equivalent to (but much slower than)
   * {@code escaper.escape(string)}: <pre>{@code
   *
   *   StringBuilder sb = new StringBuilder();
   *   escaper.escape(sb).append(string);
   *   return sb.toString();}</pre>
   *
   * @param out the underlying {@code Appendable} to append escaped output to
   * @return an {@code Appendable} which passes text to {@code out} after
   *     escaping it
   * @throws NullPointerException if {@code out} is null
   * @throws IllegalArgumentException if invalid surrogate characters are
   *         encountered
   *
   */
  public Appendable escape(final Appendable out){
    checkNotNull(out);

    return new Appendable() {

      int pendingHighSurrogate = -1;
      char[] decodedChars = new char[2];


      public Appendable append(char c) throws IOException{
        if (pendingHighSurrogate != -1) {
          // Our last append operation ended halfway through a surrogate pair
          // so we have to do some extra work first.
          if (!Character.isLowSurrogate(c)) {
            throw new IllegalArgumentException("Expected low surrogate character but got '" + c + "' with value " + (int)c);
          }
          char[] escaped = escape(Character.toCodePoint((char)pendingHighSurrogate,c));
          if (escaped != null) {
            outputChars(escaped,escaped.length);
          }
          else {
            out.append((char)pendingHighSurrogate);
            out.append(c);
          }
          pendingHighSurrogate = -1;
        }
        else if (Character.isHighSurrogate(c)) {
          // This is the start of a (split) surrogate pair.
          pendingHighSurrogate = c;
        }
        else {
          if (Character.isLowSurrogate(c)) {
            throw new IllegalArgumentException("Unexpected low surrogate character '" + c + "' with value " + (int)c);
          }
          // This is a normal (non surrogate) char.
          char[] escaped = escape(c);
          if (escaped != null) {
            outputChars(escaped,escaped.length);
          }
          else {
            out.append(c);
          }
        }
        return this;
      }


      public Appendable append(CharSequence csq) throws IOException{
        return append(csq,0,csq.length());
      }


      public Appendable append(CharSequence csq,int start,int end) throws IOException{
        int index = start;
        if (index < end) {
          // This is a little subtle: index must never reference the middle of a
          // surrogate pair but unescapedChunkStart can. The first time we enter
          // the loop below it is possible that index != unescapedChunkStart.
          int unescapedChunkStart = index;
          if (pendingHighSurrogate != -1) {
            // Our last append operation ended halfway through a surrogate pair
            // so we have to do some extra work first.
            char c = csq.charAt(index++);
            if (!Character.isLowSurrogate(c)) {
              throw new IllegalArgumentException("Expected low surrogate character but got " + c);
            }
            char[] escaped = escape(Character.toCodePoint((char)pendingHighSurrogate,c));
            if (escaped != null) {
              // Emit the escaped character and adjust unescapedChunkStart to
              // skip the low surrogate we have consumed.
              outputChars(escaped,escaped.length);
              unescapedChunkStart += 1;
            }
            else {
              // Emit pending high surrogate (unescaped) but do not modify
              // unescapedChunkStart as we must still emit the low surrogate.
              out.append((char)pendingHighSurrogate);
            }
            pendingHighSurrogate = -1;
          }
          while (true) {
            // Find and append the next subsequence of unescaped characters.
            index = nextEscapeIndex(csq,index,end);
            if (index > unescapedChunkStart) {
              out.append(csq,unescapedChunkStart,index);
            }
            if (index == end) {
              break;
            }
            // If we are not finished, calculate the next code point.
            int cp = codePointAt(csq,index,end);
            if (cp < 0) {
              // Our sequence ended half way through a surrogate pair so just
              // record the state and exit.
              pendingHighSurrogate = -cp;
              break;
            }
            // Escape the code point and output the characters.
            char[] escaped = escape(cp);
            if (escaped != null) {
              outputChars(escaped,escaped.length);
            }
            else {
              // This shouldn't really happen if nextEscapeIndex is correct but
              // we should cope with false positives.
              int len = Character.toChars(cp,decodedChars,0);
              outputChars(decodedChars,len);
            }
            // Update our index past the escaped character and continue.
            index += (Character.isSupplementaryCodePoint(cp) ? 2 : 1);
            unescapedChunkStart = index;
          }
        }
        return this;
      }


      private void outputChars(char[] chars,int len) throws IOException{
        for (int n = 0;n < len;n++) {
          out.append(chars[n]);
        }
      }
    };
  }


  /**
   * Escapes the given Unicode code point in UTF-8.
   */
  protected char[] escape(int cp){
    // We should never get negative values here but if we do it will throw an
    // IndexOutOfBoundsException, so at least it will get spotted.
    if (cp < safeOctets.length && safeOctets[cp]) {
      return null;
    }
    else if (cp == ' ' && plusForSpace) {
      return URI_ESCAPED_SPACE;
    }
    else if (cp <= 0x7F) {
      // Single byte UTF-8 characters
      // Start with "%--" and fill in the blanks
      char[] dest = new char[3];
      dest[0] = '%';
      dest[2] = UPPER_HEX_DIGITS[cp & 0xF];
      dest[1] = UPPER_HEX_DIGITS[cp >>> 4];
      return dest;
    }
    else if (cp <= 0x7ff) {
      // Two byte UTF-8 characters [cp >= 0x80 && cp <= 0x7ff]
      // Start with "%--%--" and fill in the blanks
      char[] dest = new char[6];
      dest[0] = '%';
      dest[3] = '%';
      dest[5] = UPPER_HEX_DIGITS[cp & 0xF];
      cp >>>= 4;
      dest[4] = UPPER_HEX_DIGITS[0x8 | (cp & 0x3)];
      cp >>>= 2;
      dest[2] = UPPER_HEX_DIGITS[cp & 0xF];
      cp >>>= 4;
      dest[1] = UPPER_HEX_DIGITS[0xC | cp];
      return dest;
    }
    else if (cp <= 0xffff) {
      // Three byte UTF-8 characters [cp >= 0x800 && cp <= 0xffff]
      // Start with "%E-%--%--" and fill in the blanks
      char[] dest = new char[9];
      dest[0] = '%';
      dest[1] = 'E';
      dest[3] = '%';
      dest[6] = '%';
      dest[8] = UPPER_HEX_DIGITS[cp & 0xF];
      cp >>>= 4;
      dest[7] = UPPER_HEX_DIGITS[0x8 | (cp & 0x3)];
      cp >>>= 2;
      dest[5] = UPPER_HEX_DIGITS[cp & 0xF];
      cp >>>= 4;
      dest[4] = UPPER_HEX_DIGITS[0x8 | (cp & 0x3)];
      cp >>>= 2;
      dest[2] = UPPER_HEX_DIGITS[cp];
      return dest;
    }
    else if (cp <= 0x10ffff) {
      char[] dest = new char[12];
      // Four byte UTF-8 characters [cp >= 0xffff && cp <= 0x10ffff]
      // Start with "%F-%--%--%--" and fill in the blanks
      dest[0] = '%';
      dest[1] = 'F';
      dest[3] = '%';
      dest[6] = '%';
      dest[9] = '%';
      dest[11] = UPPER_HEX_DIGITS[cp & 0xF];
      cp >>>= 4;
      dest[10] = UPPER_HEX_DIGITS[0x8 | (cp & 0x3)];
      cp >>>= 2;
      dest[8] = UPPER_HEX_DIGITS[cp & 0xF];
      cp >>>= 4;
      dest[7] = UPPER_HEX_DIGITS[0x8 | (cp & 0x3)];
      cp >>>= 2;
      dest[5] = UPPER_HEX_DIGITS[cp & 0xF];
      cp >>>= 4;
      dest[4] = UPPER_HEX_DIGITS[0x8 | (cp & 0x3)];
      cp >>>= 2;
      dest[2] = UPPER_HEX_DIGITS[cp & 0x7];
      return dest;
    }
    else {
      // If this ever happens it is due to bug in UnicodeEscaper, not bad input.
      throw new IllegalArgumentException("Invalid unicode character value " + cp);
    }
  }


  /**
   * Returns the escaped form of a given literal string.
   *
   * <p>If you are escaping input in arbitrary successive chunks, then it is not
   * generally safe to use this method. If an input string ends with an
   * unmatched high surrogate character, then this method will throw
   * {@link IllegalArgumentException}. You should either ensure your input is
   * valid <a href="http://en.wikipedia.org/wiki/UTF-16">UTF-16</a> before
   * calling this method or use an escaped {@link Appendable} (as returned by
   * {@link #escape(Appendable)}) which can cope with arbitrarily split input.
   *
   * <p><b>Note:</b> When implementing an escaper it is a good idea to override
   * this method for efficiency by inlining the implementation of
   * {@link #nextEscapeIndex(CharSequence, int, int)} directly. Doing this for
   * {@link PercentEscaper} more than doubled the performance for unescaped
   * strings (as measured by {@link CharEscapersBenchmark}).
   *
   * @param string the literal string to be escaped
   * @return the escaped form of {@code string}
   * @throws NullPointerException if {@code string} is null
   * @throws IllegalArgumentException if invalid surrogate characters are
   *         encountered
   */
  public String escape(String s){
    int slen = s.length();
    for (int index = 0;index < slen;index++) {
      char c = s.charAt(index);
      if (c >= safeOctets.length || !safeOctets[c]) {
        return escapeSlow(s,index);
      }
    }
    return s;
  }


  /**
   * Returns the escaped form of a given literal string, starting at the given
   * index.  This method is called by the {@link #escape(String)} method when it
   * discovers that escaping is required.  It is protected to allow subclasses
   * to override the fastpath escaping function to inline their escaping test.
   * See {@link CharEscaperBuilder} for an example usage.
   *
   * <p>This method is not reentrant and may only be invoked by the top level
   * {@link #escape(String)} method.
   *
   * @param s the literal string to be escaped
   * @param index the index to start escaping from
   * @return the escaped form of {@code string}
   * @throws NullPointerException if {@code string} is null
   * @throws IllegalArgumentException if invalid surrogate characters are
   *         encountered
   */
  protected final String escapeSlow(String s,int index){
    int end = s.length();

    // Get a destination buffer and setup some loop variables.
    char[] dest = DEST_TL.get();
    int destIndex = 0;
    int unescapedChunkStart = 0;

    while (index < end) {
      int cp = codePointAt(s,index,end);
      if (cp < 0) {
        throw new IllegalArgumentException("Trailing high surrogate at end of input");
      }
      char[] escaped = escape(cp);
      if (escaped != null) {
        int charsSkipped = index - unescapedChunkStart;

        // This is the size needed to add the replacement, not the full
        // size needed by the string.  We only regrow when we absolutely must.
        int sizeNeeded = destIndex + charsSkipped + escaped.length;
        if (dest.length < sizeNeeded) {
          int destLength = sizeNeeded + (end - index) + DEST_PAD;
          dest = growBuffer(dest,destIndex,destLength);
        }
        // If we have skipped any characters, we need to copy them now.
        if (charsSkipped > 0) {
          s.getChars(unescapedChunkStart,index,dest,destIndex);
          destIndex += charsSkipped;
        }
        if (escaped.length > 0) {
          System.arraycopy(escaped,0,dest,destIndex,escaped.length);
          destIndex += escaped.length;
        }
      }
      unescapedChunkStart = index + (Character.isSupplementaryCodePoint(cp) ? 2 : 1);
      index = nextEscapeIndex(s,unescapedChunkStart,end);
    }

    // Process trailing unescaped characters - no need to account for escaped
    // length or padding the allocation.
    int charsSkipped = end - unescapedChunkStart;
    if (charsSkipped > 0) {
      int endIndex = destIndex + charsSkipped;
      if (dest.length < endIndex) {
        dest = growBuffer(dest,destIndex,endIndex);
      }
      s.getChars(unescapedChunkStart,end,dest,destIndex);
      destIndex = endIndex;
    }
    return new String(dest,0,destIndex);
  }


  /**
   * Scans a sub-sequence of characters from a given {@link CharSequence},
   * returning the index of the next character that requires escaping.
   *
   * <p><b>Note:</b> When implementing an escaper, it is a good idea to override
   * this method for efficiency. The base class implementation determines
   * successive Unicode code points and invokes {@link #escape(int)} for each of
   * them. If the semantics of your escaper are such that code points in the
   * supplementary range are either all escaped or all unescaped, this method
   * can be implemented more efficiently using {@link CharSequence#charAt(int)}.
   *
   * <p>Note however that if your escaper does not escape characters in the
   * supplementary range, you should either continue to validate the correctness
   * of any surrogate characters encountered or provide a clear warning to users
   * that your escaper does not validate its input.
   *
   * <p>See {@link PercentEscaper} for an example.
   *
   * @param csq a sequence of characters
   * @param start the index of the first character to be scanned
   * @param end the index immediately after the last character to be scanned
   * @throws IllegalArgumentException if the scanned sub-sequence of {@code csq}
   *     contains invalid surrogate pairs
   */
  protected int nextEscapeIndex(CharSequence csq,int index,int end){
    for (;index < end;index++) {
      char c = csq.charAt(index);
      if (c >= safeOctets.length || !safeOctets[c]) {
        break;
      }
    }
    return index;
  }


  /**
   * Ensures the truth of an expression involving one or more parameters to the
   * calling method.
   *
   * @param expression a boolean expression
   * @throws IllegalArgumentException if {@code expression} is false
   */
  public static void checkArgument(boolean expression){
    if (!expression) {
      throw new IllegalArgumentException();
    }
  }


  /**
   * Ensures the truth of an expression involving one or more parameters to the
   * calling method.
   *
   * @param expression a boolean expression
   * @param errorMessage the exception message to use if the check fails; will
   *     be converted to a string using {@link String#valueOf(Object)}
   * @throws IllegalArgumentException if {@code expression} is false
   */
  public static void checkArgument(boolean expression,Object errorMessage){
    if (!expression) {
      throw new IllegalArgumentException(String.valueOf(errorMessage));
    }
  }


  /**
   * Ensures the truth of an expression involving one or more parameters to the
   * calling method.
   *
   * @param expression a boolean expression
   * @param errorMessageTemplate a template for the exception message should the
   *     check fail. The message is formed by replacing each {@code %s}
   *     placeholder in the template with an argument. These are matched by
   *     position - the first {@code %s} gets {@code errorMessageArgs[0]}, etc.
   *     Unmatched arguments will be appended to the formatted message in square
   *     braces. Unmatched placeholders will be left as-is.
   * @param errorMessageArgs the arguments to be substituted into the message
   *     template. Arguments are converted to strings using
   *     {@link String#valueOf(Object)}.
   * @throws IllegalArgumentException if {@code expression} is false
   * @throws NullPointerException if the check fails and either {@code
   *     errorMessageTemplate} or {@code errorMessageArgs} is null (don't let
   *     this happen)
   */
  public static void checkArgument(boolean expression,String errorMessageTemplate,Object ... errorMessageArgs){
    if (!expression) {
      throw new IllegalArgumentException(format(errorMessageTemplate,errorMessageArgs));
    }
  }


  /**
   * Ensures that an {@code Iterable} object passed as a parameter to the
   * calling method is not null and contains no null elements.
   *
   * @param iterable the iterable to check the contents of
   * @return the non-null {@code iterable} reference just validated
   * @throws NullPointerException if {@code iterable} is null or contains at
   *     least one null element
   */
  public static <T extends Iterable<?>> T checkContentsNotNull(T iterable){
    if (containsOrIsNull(iterable)) {
      throw new NullPointerException();
    }
    return iterable;
  }


  /**
   * Ensures that an {@code Iterable} object passed as a parameter to the
   * calling method is not null and contains no null elements.
   *
   * @param iterable the iterable to check the contents of
   * @param errorMessage the exception message to use if the check fails; will
   *     be converted to a string using {@link String#valueOf(Object)}
   * @return the non-null {@code iterable} reference just validated
   * @throws NullPointerException if {@code iterable} is null or contains at
   *     least one null element
   */
  public static <T extends Iterable<?>> T checkContentsNotNull(T iterable,Object errorMessage){
    if (containsOrIsNull(iterable)) {
      throw new NullPointerException(String.valueOf(errorMessage));
    }
    return iterable;
  }


  /**
   * Ensures that an {@code Iterable} object passed as a parameter to the
   * calling method is not null and contains no null elements.
   *
   * @param iterable the iterable to check the contents of
   * @param errorMessageTemplate a template for the exception message should the
   *     check fail. The message is formed by replacing each {@code %s}
   *     placeholder in the template with an argument. These are matched by
   *     position - the first {@code %s} gets {@code errorMessageArgs[0]}, etc.
   *     Unmatched arguments will be appended to the formatted message in square
   *     braces. Unmatched placeholders will be left as-is.
   * @param errorMessageArgs the arguments to be substituted into the message
   *     template. Arguments are converted to strings using
   *     {@link String#valueOf(Object)}.
   * @return the non-null {@code iterable} reference just validated
   * @throws NullPointerException if {@code iterable} is null or contains at
   *     least one null element
   */
  public static <T extends Iterable<?>> T checkContentsNotNull(T iterable,String errorMessageTemplate,Object ... errorMessageArgs){
    if (containsOrIsNull(iterable)) {
      throw new NullPointerException(format(errorMessageTemplate,errorMessageArgs));
    }
    return iterable;
  }


  /**
   * Ensures that {@code index} specifies a valid <i>element</i> in an array,
   * list or string of size {@code size}. An element index may range from zero,
   * inclusive, to {@code size}, exclusive.
   *
   * @param index a user-supplied index identifying an element of an array, list
   *     or string
   * @param size the size of that array, list or string
   * @throws IndexOutOfBoundsException if {@code index} is negative or is not
   *     less than {@code size}
   * @throws IllegalArgumentException if {@code size} is negative
   */
  public static void checkElementIndex(int index,int size){
    checkElementIndex(index,size,"index");
  }


  /**
   * Ensures that {@code index} specifies a valid <i>element</i> in an array,
   * list or string of size {@code size}. An element index may range from zero,
   * inclusive, to {@code size}, exclusive.
   *
   * @param index a user-supplied index identifying an element of an array, list
   *     or string
   * @param size the size of that array, list or string
   * @param desc the text to use to describe this index in an error message
   * @throws IndexOutOfBoundsException if {@code index} is negative or is not
   *     less than {@code size}
   * @throws IllegalArgumentException if {@code size} is negative
   */
  public static void checkElementIndex(int index,int size,String desc){
    checkArgument(size >= 0,"negative size: %s",size);
    if (index < 0) {
      throw new IndexOutOfBoundsException(format("%s (%s) must not be negative",desc,index));
    }
    if (index >= size) {
      throw new IndexOutOfBoundsException(format("%s (%s) must be less than size (%s)",desc,index,size));
    }
  }


  /**
   * Ensures that an object reference passed as a parameter to the calling
   * method is not null.
   *
   * @param reference an object reference
   * @return the non-null reference that was validated
   * @throws NullPointerException if {@code reference} is null
   */
  public static <T> T checkNotNull(T reference){
    if (reference == null) {
      throw new NullPointerException();
    }
    return reference;
  }


  /**
   * Ensures that an object reference passed as a parameter to the calling
   * method is not null.
   *
   * @param reference an object reference
   * @param errorMessage the exception message to use if the check fails; will
   *     be converted to a string using {@link String#valueOf(Object)}
   * @return the non-null reference that was validated
   * @throws NullPointerException if {@code reference} is null
   */
  public static <T> T checkNotNull(T reference,Object errorMessage){
    if (reference == null) {
      throw new NullPointerException(String.valueOf(errorMessage));
    }
    return reference;
  }


  /**
   * Ensures that an object reference passed as a parameter to the calling
   * method is not null.
   *
   * @param reference an object reference
   * @param errorMessageTemplate a template for the exception message should the
   *     check fail. The message is formed by replacing each {@code %s}
   *     placeholder in the template with an argument. These are matched by
   *     position - the first {@code %s} gets {@code errorMessageArgs[0]}, etc.
   *     Unmatched arguments will be appended to the formatted message in square
   *     braces. Unmatched placeholders will be left as-is.
   * @param errorMessageArgs the arguments to be substituted into the message
   *     template. Arguments are converted to strings using
   *     {@link String#valueOf(Object)}.
   * @return the non-null reference that was validated
   * @throws NullPointerException if {@code reference} is null
   */
  public static <T> T checkNotNull(T reference,String errorMessageTemplate,Object ... errorMessageArgs){
    if (reference == null) {
      // If either of these parameters is null, the right thing happens anyway
      throw new NullPointerException(format(errorMessageTemplate,errorMessageArgs));
    }
    return reference;
  }


  /**
   * Ensures that {@code index} specifies a valid <i>position</i> in an array,
   * list or string of size {@code size}. A position index may range from zero
   * to {@code size}, inclusive.
   *
   * @param index a user-supplied index identifying a position in an array, list
   *     or string
   * @param size the size of that array, list or string
   * @throws IndexOutOfBoundsException if {@code index} is negative or is
   *     greater than {@code size}
   * @throws IllegalArgumentException if {@code size} is negative
   */
  public static void checkPositionIndex(int index,int size){
    checkPositionIndex(index,size,"index");
  }


  /**
   * Ensures that {@code index} specifies a valid <i>position</i> in an array,
   * list or string of size {@code size}. A position index may range from zero
   * to {@code size}, inclusive.
   *
   * @param index a user-supplied index identifying a position in an array, list
   *     or string
   * @param size the size of that array, list or string
   * @param desc the text to use to describe this index in an error message
   * @throws IndexOutOfBoundsException if {@code index} is negative or is
   *     greater than {@code size}
   * @throws IllegalArgumentException if {@code size} is negative
   */
  public static void checkPositionIndex(int index,int size,String desc){
    checkArgument(size >= 0,"negative size: %s",size);
    if (index < 0) {
      throw new IndexOutOfBoundsException(format("%s (%s) must not be negative",desc,index));
    }
    if (index > size) {
      throw new IndexOutOfBoundsException(format("%s (%s) must not be greater than size (%s)",desc,index,size));
    }
  }


  /**
   * Ensures that {@code start} and {@code end} specify a valid <i>positions</i>
   * in an array, list or string of size {@code size}, and are in order. A
   * position index may range from zero to {@code size}, inclusive.
   *
   * @param start a user-supplied index identifying a starting position in an
   *     array, list or string
   * @param end a user-supplied index identifying a ending position in an array,
   *     list or string
   * @param size the size of that array, list or string
   * @throws IndexOutOfBoundsException if either index is negative or is
   *     greater than {@code size}, or if {@code end} is less than {@code start}
   * @throws IllegalArgumentException if {@code size} is negative
   */
  public static void checkPositionIndexes(int start,int end,int size){
    checkPositionIndex(start,size,"start index");
    checkPositionIndex(end,size,"end index");
    if (end < start) {
      throw new IndexOutOfBoundsException(format("end index (%s) must not be less than start index (%s)",end,start));
    }
  }


  /**
   * Ensures the truth of an expression involving the state of the calling
   * instance, but not involving any parameters to the calling method.
   *
   * @param expression a boolean expression
   * @throws IllegalStateException if {@code expression} is false
   */
  public static void checkState(boolean expression){
    if (!expression) {
      throw new IllegalStateException();
    }
  }


  /**
   * Ensures the truth of an expression involving the state of the calling
   * instance, but not involving any parameters to the calling method.
   *
   * @param expression a boolean expression
   * @param errorMessage the exception message to use if the check fails; will
   *     be converted to a string using {@link String#valueOf(Object)}
   * @throws IllegalStateException if {@code expression} is false
   */
  public static void checkState(boolean expression,Object errorMessage){
    if (!expression) {
      throw new IllegalStateException(String.valueOf(errorMessage));
    }
  }


  /**
   * Ensures the truth of an expression involving the state of the calling
   * instance, but not involving any parameters to the calling method.
   *
   * @param expression a boolean expression
   * @param errorMessageTemplate a template for the exception message should the
   *     check fail. The message is formed by replacing each {@code %s}
   *     placeholder in the template with an argument. These are matched by
   *     position - the first {@code %s} gets {@code errorMessageArgs[0]}, etc.
   *     Unmatched arguments will be appended to the formatted message in square
   *     braces. Unmatched placeholders will be left as-is.
   * @param errorMessageArgs the arguments to be substituted into the message
   *     template. Arguments are converted to strings using
   *     {@link String#valueOf(Object)}.
   * @throws IllegalStateException if {@code expression} is false
   * @throws NullPointerException if the check fails and either {@code
   *     errorMessageTemplate} or {@code errorMessageArgs} is null (don't let
   *     this happen)
   */
  public static void checkState(boolean expression,String errorMessageTemplate,Object ... errorMessageArgs){
    if (!expression) {
      throw new IllegalStateException(format(errorMessageTemplate,errorMessageArgs));
    }
  }


  /**
   * Returns the Unicode code point of the character at the given index.
   *
   * <p>Unlike {@link Character#codePointAt(CharSequence, int)} or
   * {@link String#codePointAt(int)} this method will never fail silently when
   * encountering an invalid surrogate pair.
   *
   * <p>The behaviour of this method is as follows:
   * <ol>
   * <li>If {@code index >= end}, {@link IndexOutOfBoundsException} is thrown.
   * <li><b>If the character at the specified index is not a surrogate, it is
   *     returned.</b>
   * <li>If the first character was a high surrogate value, then an attempt is
   *     made to read the next character.
   *     <ol>
   *     <li><b>If the end of the sequence was reached, the negated value of
   *         the trailing high surrogate is returned.</b>
   *     <li><b>If the next character was a valid low surrogate, the code point
   *         value of the high/low surrogate pair is returned.</b>
   *     <li>If the next character was not a low surrogate value, then
   *         {@link IllegalArgumentException} is thrown.
   *     </ol>
   * <li>If the first character was a low surrogate value,
   *     {@link IllegalArgumentException} is thrown.
   * </ol>
   *
   * @param seq the sequence of characters from which to decode the code point
   * @param index the index of the first character to decode
   * @param end the index beyond the last valid character to decode
   * @return the Unicode code point for the given index or the negated value of
   *         the trailing high surrogate character at the end of the sequence
   */
  protected static final int codePointAt(CharSequence seq,int index,int end){
    if (index < end) {
      char c1 = seq.charAt(index++);
      if (c1 < Character.MIN_HIGH_SURROGATE || c1 > Character.MAX_LOW_SURROGATE) {
        // Fast path (first test is probably all we need to do)
        return c1;
      }
      else if (c1 <= Character.MAX_HIGH_SURROGATE) {
        // If the high surrogate was the last character, return its inverse
        if (index == end) {
          return -c1;
        }
        // Otherwise look for the low surrogate following it
        char c2 = seq.charAt(index);
        if (Character.isLowSurrogate(c2)) {
          return Character.toCodePoint(c1,c2);
        }
        throw new IllegalArgumentException("Expected low surrogate but got char '" + c2 + "' with value " + (int)c2 + " at index " + index);
      }
      else {
        throw new IllegalArgumentException("Unexpected low surrogate character '" + c1 + "' with value " + (int)c1 + " at index " + (index - 1));
      }
    }
    throw new IndexOutOfBoundsException("Index exceeds specified range");
  }


  private static boolean containsOrIsNull(Iterable<?> iterable){
    if (iterable == null) {
      return true;
    }

    if (iterable instanceof Collection) {
      Collection<?> collection = (Collection<?>)iterable;
      try {
        return collection.contains(null);
      }
      catch (NullPointerException e) {
        // A NPE implies that the collection doesn't contain null.
        return false;
      }
    }
    else {
      for (Object element : iterable) {
        if (element == null) {
          return true;
        }
      }
      return false;
    }
  }


  /**
   * Creates a boolean[] with entries corresponding to the character values
   * for 0-9, A-Z, a-z and those specified in safeChars set to true. The array
   * is as small as is required to hold the given character information.
   */
  private static boolean[] createSafeOctets(String safeChars){
    int maxChar = 'z';
    char[] safeCharArray = safeChars.toCharArray();
    for (char c : safeCharArray) {
      maxChar = Math.max(c,maxChar);
    }
    boolean[] octets = new boolean[maxChar + 1];
    for (int c = '0';c <= '9';c++) {
      octets[c] = true;
    }
    for (int c = 'A';c <= 'Z';c++) {
      octets[c] = true;
    }
    for (int c = 'a';c <= 'z';c++) {
      octets[c] = true;
    }
    for (char c : safeCharArray) {
      octets[c] = true;
    }
    return octets;
  }


  /**
   * Substitutes each {@code %s} in {@code template} with an argument. These
   * are matched by position - the first {@code %s} gets {@code args[0]}, etc.
   * If there are more arguments than placeholders, the unmatched arguments will
   * be appended to the end of the formatted message in square braces.
   *
   * @param template a non-null string containing 0 or more {@code %s}
   *     placeholders.
   * @param args the arguments to be substituted into the message
   *     template. Arguments are converted to strings using
   *     {@link String#valueOf(Object)}. Arguments can be null.
   */
  // VisibleForTesting
  static String format(String template,Object ... args){
    // start substituting the arguments into the '%s' placeholders
    StringBuilder builder = new StringBuilder(template.length() + 16 * args.length);
    int templateStart = 0;
    int i = 0;
    while (i < args.length) {
      int placeholderStart = template.indexOf("%s",templateStart);
      if (placeholderStart == -1) {
        break;
      }
      builder.append(template.substring(templateStart,placeholderStart));
      builder.append(args[i++]);
      templateStart = placeholderStart + 2;
    }
    builder.append(template.substring(templateStart));

    // if we run out of placeholders, append the extra args in square braces
    if (i < args.length) {
      builder.append(" [");
      builder.append(args[i++]);
      while (i < args.length) {
        builder.append(", ");
        builder.append(args[i++]);
      }
      builder.append("]");
    }

    return builder.toString();
  }


  /**
   * Helper method to grow the character buffer as needed, this only happens
   * once in a while so it's ok if it's in a method call.  If the index passed
   * in is 0 then no copying will be done.
   */
  private static final char[] growBuffer(char[] dest,int index,int size){
    char[] copy = new char[size];
    if (index > 0) {
      System.arraycopy(dest,0,copy,0,index);
    }
    return copy;
  }
}
