////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                            //
//                                       WATTO STUDIOS                                        //
//                             Java Code, Programs, and Software                              //
//                                    http://www.watto.org                                    //
//                                                                                            //
//                           Copyright (C) 2004-2010  WATTO Studios                           //
//                                                                                            //
// This program is free software; you can redistribute it and/or modify it under the terms of //
// the GNU General Public License published by the Free Software Foundation; either version 2 //
// of the License, or (at your option) any later versions. This program is distributed in the //
// hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranties //
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License //
// at http://www.gnu.org for more details. For updates and information about this program, go //
// to the WATTO Studios website at http://www.watto.org or email watto@watto.org . Thanks! :) //
//                                                                                            //
////////////////////////////////////////////////////////////////////////////////////////////////

package org.watto.program.android.sync.miso.security;

import java.io.IOException;
import java.net.URLEncoder;
import java.util.Arrays;
import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import org.apache.commons.codec.binary.CustomBase64;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.ParseException;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.scheme.PlainSocketFactory;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.tsccm.ThreadSafeClientConnManager;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.HttpParams;
import org.apache.http.util.EntityUtils;
import org.watto.program.android.sync.miso.PercentEscaper;


/***********************************************************************************************
Convenience methods for establishing OAuth 2.0 authentication to Miso
***********************************************************************************************/
public class MisoOAuth1Client extends OAuth1Client {

  /***********************************************************************************************
  Initialises the OAuth provider with the <i>consumerKey</i>, <i>consumerSecret</i> and the 
  <i>callbackUrl</i> for the application
  @param consumerKey the OAuth key for this application
  @param consumerSecret the OAuth secret for this application
  @param callbackUrl the OAuth callback URL for this application
  ***********************************************************************************************/
  public MisoOAuth1Client(String consumerKey, String consumerSecret, String callbackUrl){
    super(consumerKey,consumerSecret,callbackUrl);
  }


  /***********************************************************************************************
  Gets the URL for converting an authentication <i>code</i> into an access token
  @param code the authentication code
  @return the URL used to generate the access token
  ***********************************************************************************************/
  public String getRequestTokenUrl(){
    String encodedCallbackUrl = URLEncoder.encode(callbackUrl);
    String timestamp = Long.toString(System.currentTimeMillis() / 1000L);
    String nonce = "requesttoken" + timestamp; 
    String url = "http://gomiso.com/oauth/request_token?oauth_consumer_key=" + consumerKey + "&oauth_nonce=" + nonce + "&oauth_signature_method=HMAC-SHA1&oauth_timestamp=" + timestamp + "&oauth_version=1.0&oauth_callback=" + encodedCallbackUrl;
    return signRequest(url);
  }
  
  
  /***********************************************************************************************
  Adds the additional OAuth parameters such as timestamp, nonce, signature_method, and version
  ***********************************************************************************************/
  public String addExtraOAuthParams(String url){

    String timestamp = Long.toString(System.currentTimeMillis() / 1000L);
    String nonce = "request" + timestamp; 
    
    String extraParams = "oauth_consumer_key=" + consumerKey + "&oauth_consumer_secret=" + consumerSecret + "&oauth_nonce=" + nonce + "&oauth_signature_method=HMAC-SHA1&oauth_timestamp=" + timestamp + "&oauth_version=1.0";
    
      if (url.indexOf('?') > 0) {
        // already has a '?', so need a '&' instead
        url += "&" + extraParams;
      }
      else {
        url += "?" + extraParams;
      }
      
      return url;
  }


  /***********************************************************************************************
  Gets the URL for converting an <i>authorisationToken</i> and <i>verifier</i> into an access token
  @param authorisationToken the authorisation token
  @param verifier the authorisation verifier
  @return the URL used to generate the access token
  ***********************************************************************************************/
  public String getAccessTokenUrl(String authorisationToken,String verifier){
    String timestamp = Long.toString(System.currentTimeMillis() / 1000L);
    String url = "http://gomiso.com/oauth/access_token?oauth_consumer_key=" + consumerKey + "&oauth_nonce=accesstokenrequest&oauth_signature_method=HMAC-SHA1&oauth_timestamp=" + timestamp + "&oauth_token=" + authorisationToken + "&oauth_verifier=" + verifier + "&oauth_version=1.0";
    return signRequest(url);
  }


  /***********************************************************************************************
  Gets the URL for authorizing the application
  @return the URL used to authorize the application
  ***********************************************************************************************/
  public String getAuthorizationUrl(){
    String requestToken = getRequestToken();
    if(requestToken == null || requestToken.equals("")){
      return null;
    }
    return "http://gomiso.com/oauth/authorize?oauth_token=" + requestToken + "&permission=read";
  }


  /***********************************************************************************************
  Gets the request token
  @return the request token
  ***********************************************************************************************/
  public String retreiveRequestToken(){
    String tokenURL = getRequestTokenUrl();

    // Build a HttpClient that ignores SSL errors
    SchemeRegistry schemeRegistry = new SchemeRegistry();
    schemeRegistry.register(new Scheme("http",PlainSocketFactory.getSocketFactory(),80));
    schemeRegistry.register(new Scheme("https",new EasySSLSocketFactory(),443));

    HttpParams params = new BasicHttpParams();
    ClientConnectionManager clientConnectionManager = new ThreadSafeClientConnManager(params,schemeRegistry);

    DefaultHttpClient client = new DefaultHttpClient(clientConnectionManager,params);
    HttpGet get = new HttpGet(tokenURL);

    // Call the URL
    HttpResponse response = null;
    try {
      response = client.execute(get);
    }
    catch (ClientProtocolException e) {}
    catch (IOException e) {}

    String token = "";
    String tokenSecret = "";
    if (response == null || response.getStatusLine().getStatusCode() != HttpStatus.SC_OK) {
      // error
    }
    else {
      // success - got a token
      try {
        // eg oauth_token=8uHLkN4P9UNpsT1E0Ug&oauth_token_secret=2E6ipNmatz4SWbNsOn86Sd30Kk6wb3oRd25CiNuC&oauth_callback_confirmed=true
        String content = EntityUtils.toString(response.getEntity());
        
        String[] parameters = content.split("&");
        int parameterCount = parameters.length;
        
        for (int i=0;i<parameterCount;i++){
          String[] splitPair = parameters[i].split("=");
          if (splitPair[0].equals("oauth_token")){
            token = splitPair[1];
            continue;
          }
          else if (splitPair[0].equals("oauth_token_secret")){
            tokenSecret = splitPair[1];
            continue;
          }
        }
      }
      catch (ParseException e) {}
      catch (IOException e) {
        System.out.println(e);
      }
    }

    setRequestToken(token);
    setRequestTokenSecret(tokenSecret);
    return token;

  }


  /***********************************************************************************************
  Converts the <i>authorisationToken</i> and <i>verifier</i> into an access token
  @param authToken the authorisation token
  @param verifier the authorisation verifier
  @return the access token
  ***********************************************************************************************/
  public String retreiveAccessToken(String authToken, String verifier){
    String tokenURL = getAccessTokenUrl(authToken,verifier);

    // Build a HttpClient that ignores SSL errors
    SchemeRegistry schemeRegistry = new SchemeRegistry();
    schemeRegistry.register(new Scheme("http",PlainSocketFactory.getSocketFactory(),80));
    schemeRegistry.register(new Scheme("https",new EasySSLSocketFactory(),443));

    HttpParams params = new BasicHttpParams();
    ClientConnectionManager clientConnectionManager = new ThreadSafeClientConnManager(params,schemeRegistry);

    DefaultHttpClient client = new DefaultHttpClient(clientConnectionManager,params);
    HttpGet get = new HttpGet(tokenURL);

    // Call the URL
    HttpResponse response = null;
    try {
      response = client.execute(get);
    }
    catch (ClientProtocolException e) {}
    catch (IOException e) {}

    String token = "";
    String tokenSecret = "";
    if (response == null || response.getStatusLine().getStatusCode() != HttpStatus.SC_OK) {
      // error
    }
    else {
      // success - got a token
      try {
        String content = EntityUtils.toString(response.getEntity());
        
        String[] parameters = content.split("&");
        int parameterCount = parameters.length;
        
        for (int i=0;i<parameterCount;i++){
          String[] splitPair = parameters[i].split("=");
          if (splitPair[0].equals("oauth_token")){
            token = splitPair[1];
            continue;
          }
          else if (splitPair[0].equals("oauth_token_secret")){
            tokenSecret = splitPair[1];
            continue;
          }
        }
      }
      catch (ParseException e) {}
      catch (IOException e) {
        System.out.println(e);
      }
    }

    setAccessToken(token);
    setAccessTokenSecret(tokenSecret);
    return token;

  }
  
  
  /***********************************************************************************************

  ***********************************************************************************************/
  public String percentEncode(String message){
    return percentEscaper.escape(message);
  }


  /***********************************************************************************************
  Signs a URL request with the access token
  @param url the request to sign
  @return the signed request
  ***********************************************************************************************/
  public String signRequest(String url){
    // Add the access token, if we have it yet
    String accessToken = getAccessToken();
    if (accessToken != null && !accessToken.equals("")) {
      String accessTokenSecret = getAccessTokenSecret();
      
      url = addExtraOAuthParams(url);
      
      if (url.indexOf('?') > 0) {
        // already has a '?', so need a '&' instead
        url += "&oauth_token=" + accessToken + "&oauth_token_secret=" + accessTokenSecret;
      }
      else {
        url += "?oauth_token=" + accessToken + "&oauth_token_secret=" + accessTokenSecret;
      }
    }
    
    
    



      // Add the signature
      String signature = calculateSHA1(url);
      int generationCount = 0;
      while (signature.indexOf('+') >= 0 && generationCount < 10){
        // Detected a '+' sign in the signature, which will fail as a 401 Unauthorised.
        // So, if there is a '+' sign, change the oauth_nonce value and try again.
        // We give up after 10 attempts!
        
        // Find the nonce value
        int noncePosition = url.indexOf("oauth_nonce=") + 12;
        int nonceEnd = url.indexOf("&",noncePosition);
        
        // Generate a new nonce value
        String newNonce = "retry" + Math.random()*10000;
        
        // Replace the nonce value
        url = url.substring(0,noncePosition) + newNonce + url.substring(nonceEnd);
        
        // Re-generate the signature
        signature = calculateSHA1(url);
        
        generationCount++;
      }
      
      
      if (url.indexOf('?') > 0) {
        // already has a '?', so need a '&' instead
        url += "&oauth_signature=" + signature;
      }
      else {
        url += "?oauth_signature=" + signature;
      }
      


      return url;

  }

  PercentEscaper percentEscaper = null;

  /***********************************************************************************************
  Calculates the SHA1 hash for a <i>url</i> and returns it in Hex format
  @param url the <code>String</code> to calculate the hash for
  @return the SHA1 hash in Base64 format
  ***********************************************************************************************/
  public String calculateSHA1(String url){
    
    if (percentEscaper == null){
      percentEscaper = new PercentEscaper("-._",false);
    }
    
    try {
      
      // 1. Split the url into the address and the parameters
      String address = url;
      String parameters = "";
      
      int questionPosition = url.indexOf('?');
      if (questionPosition > 0){
        address = url.substring(0,questionPosition);
        parameters = url.substring(questionPosition+1);
      }
      
      // 2. Split the parameters up into key/value pairs and sort them alphabetically
      String[] parameterPairs = parameters.split("&");
      Arrays.sort(parameterPairs);
      
      // 3. Build the url used in the signature
      String parametersRebuilt = "";
      int parameterCount = parameterPairs.length;
      for (int i=0;i<parameterCount;i++){
        if (i != 0){
          parametersRebuilt += "&";
        }
        parametersRebuilt += parameterPairs[i];
      }
      String encodedUrl = "GET&" + percentEncode(address) + "&" + percentEncode(parametersRebuilt);
      
      

      
  
      

      String key = getConsumerSecret() + "&" + getAccessTokenSecret();//getRequestTokenSecret();

      // Create the SHA1 Hash
      SecretKeySpec keySpec = new SecretKeySpec(key.getBytes("UTF-8"),"HmacSHA1");
      Mac mac = Mac.getInstance("HmacSHA1");
      mac.init(keySpec);

      byte[] bytes = mac.doFinal(encodedUrl.getBytes("UTF-8"));

      // Convert to Base64
      CustomBase64 base64 = new CustomBase64();
      return new String(base64.encode(bytes)).trim();

    }
    catch (Throwable t) {
      return "";
    }
  }

}
